from __future__ import annotations

from datetime import datetime
import re
from typing import Dict, List, Optional

from .rule import Rule

aiohttp_installed = False
requests_installed = False

try:
    import aiohttp
    aiohttp_installed = True
except ImportError:
    pass

try:
    import requests
    requests_installed = True
except ImportError:
    pass


if not aiohttp_installed and not requests_installed:
    raise ImportError("At least one web driver is required. Please install aiohttp or requests.")


class AnimeRelations:
    session: Optional[aiohttp.ClientSession]

    username: str
    repo_name: str
    branch_name: str
    file_path: str

    last_fetch: Optional[datetime]
    meta: Dict[str, str]
    rules: List[Rule]

    def __init__(self, **kwargs) -> None:
        """
        Creates a new AnimeRelations parser with the
        passed parameters.

        Parameters
        ----------
        session: Optional[aiohttp.ClientSession]
            Session for retrieving the file(s).
        username: str
            The username the repo is hosted
            under.
        repo_name: str
            The repository the file is hosted
            in.
        branch_name: str
            The branch that the file is located
            in.
        file_path: str
            The name of the file that contains
            the rules.
        """
        self.session = kwargs.pop("session", None)
        if aiohttp_installed and self.session is None:
            self.session = aiohttp.ClientSession()

        self.username = kwargs.pop("username", "erengy")
        self.repo_name = kwargs.pop("repo_name", "anime-relations")
        self.branch_name = kwargs.pop("branch_name", "master")
        self.file_path = kwargs.pop("file_path", "anime-relations.txt")

        self.last_fetch = None
        self.meta = {}
        self.rules = []

    async def fetch_async(self) -> None:
        """
        Asynchronously fetches new rules with the
        existing configuration.

        Updates the instance with the parsed new rules.
        """
        if self.session is None:
            raise RuntimeError("Cannot fetch async if aiohttp is not installed.")

        url = f"https://raw.githubusercontent.com/{self.username}/{self.repo_name}/{self.branch_name}/{self.file_path}"
        async with self.session.get(url) as resp:
            self.parse(await resp.text())

        self.last_fetch = datetime.utcnow()

    def fetch_sync(self) -> None:
        """
        Synchronously fetches new rules with the
        existing configuration.

        Updates the instance with the parsed new rules.
        """
        if not requests_installed:
            raise RuntimeError("Cannot fetch sync if requests is not installed.")

        url = f"https://raw.githubusercontent.com/{self.username}/{self.repo_name}/{self.branch_name}/{self.file_path}"
        resp = requests.get(url)
        self.parse(resp.text)

        self.last_fetch = datetime.utcnow()

    def parse(self, data: str) -> None:
        """
        Parses a string of data for rules and meta.

        Parameters
        ----------
        data: str
            String to parse for rules and meta.
        """
        in_meta = False
        in_rules = False

        for line in data.split("\n"):
            if line.startswith("#"):
                continue
            elif line.startswith("::meta"):
                in_meta = True
                in_rules = False
            elif line.startswith("::rules"):
                in_rules = True
                in_meta = False

            if in_meta:
                rule = re.match(r"- (\S+): (\S+)", line)
                if rule is None:
                    continue
                self.meta[rule.group(1)] = rule.group(2)
            elif in_rules:
                rules = Rule.from_line(line)
                self.rules += rules

    def from_mal(self, mal_id: int) -> Optional[Rule]:
        """
        Retrieves a Rule based on the passed MAL id.

        Parameters
        ----------
        mal_id: int
            The rule with the MAL id to
            search for.

        Returns
        -------
        Optional[Rule]
            The found rule.
        """
        for rule in self.rules:
            if rule.mal_from == mal_id:
                return rule

    def from_kitsu(self, kitsu_id: int) -> Optional[Rule]:
        """
        Retrieves a Rule based on the passed kitsu id.

        Parameters
        ----------
        kitsu_id: int
            The rule with the kitsu id to
            search for.

        Returns
        -------
        Optional[Rule]
            The found rule.
        """
        for rule in self.rules:
            if rule.kitsu_from == kitsu_id:
                return rule

    def from_anilist(self, anilist_id: int) -> Optional[Rule]:
        """
        Retrieves a Rule based on the passed anilist id.

        Parameters
        ----------
        anilist_id: int
            The rule with the anilist id to
            search for.

        Returns
        -------
        Optional[Rule]
            The found rule.
        """
        for rule in self.rules:
            if rule.anilist_from == anilist_id:
                return rule



from __future__ import annotations

from dataclasses import dataclass
import re
from typing import List, Optional, Tuple


RULE_PATTERN = re.compile(r"- (?P<mal_from>[\d\?]+)\|(?P<kitsu_from>[\d\?]+)\|(?P<anilist_from>[\d\?]+):(?P<episodes_from>[\d\-]+) -> (?P<mal_to>[\d\?~]+)\|(?P<kitsu_to>[\d\?~]+)\|(?P<anilist_to>[\d\?~]+):(?P<episodes_to>[\d\-!]+)")


@dataclass
class Rule:
    raw: str

    mal_from: Optional[int]
    kitsu_from: Optional[int]
    anilist_from: Optional[int]

    episodes_from: Tuple[int, Optional[int]]

    mal_to: Optional[int]
    kitsu_to: Optional[int]
    anilist_to: Optional[int]

    episodes_to: Tuple[int, Optional[int]]

    @classmethod
    def from_line(cls, line: str) -> List[Rule]:
        """
        Creates a Rule instance from a line
        of text.

        Returns a list of rules, as it is possible
        that multiple rules can be created from a
        single line.

        If there is not a valid rule in the passed
        string, then nothing will be returned.

        Parameters
        ----------
        line: str
            The line to find a rule in.

        Returns
        -------
        List[Rule]
            The rules.
        """
        instances = []
        parsed = re.match(RULE_PATTERN, line)
        if parsed is None:
            return instances

        episodes_from = parsed.group("episodes_from")
        if "-" in episodes_from:
            start, end = episodes_from.split("-")
            if end != "?":
                episodes_from = (int(start), int(end))
            else:
                episodes_from = (int(start), None)
        else:
            episodes_from = (int(episodes_from), int(episodes_from))

        mal_from = parsed.group("mal_from")
        mal_from = int(mal_from) if mal_from != "?" else None

        kitsu_from = parsed.group("kitsu_from")
        kitsu_from = int(kitsu_from) if kitsu_from != "?" else None

        anilist_from = parsed.group("anilist_from")
        anilist_from = int(anilist_from) if anilist_from != "?" else None

        mal_to = parsed.group("mal_to")
        if mal_to == "~":
            mal_to = mal_from
        else:
            mal_to = int(mal_to) if mal_to != "?" else None

        kitsu_to = parsed.group("kitsu_to")
        if kitsu_to == "~":
            kitsu_to = kitsu_from
        else:
            kitsu_to = int(kitsu_to) if kitsu_to != "?" else None

        anilist_to = parsed.group("anilist_to")
        if anilist_to == "~":
            anilist_to = anilist_from
        else:
            anilist_to = int(anilist_to) if anilist_to != "?" else None

        episodes_to = parsed.group("episodes_to").replace("!", "")
        if "-" in episodes_to:
            start, end = episodes_to.split("-")
            if end != "?":
                episodes_to = (int(start), int(end))
            else:
                episodes_to = (int(start), None)
        else:
            episodes_to = (int(episodes_to), int(episodes_to))

        if parsed.group("episodes_to").endswith("!"):
            redirect = cls(
                raw=parsed.group(0),
                mal_from=mal_to,
                kitsu_from=kitsu_to,
                anilist_from=anilist_to,
                episodes_from=episodes_from,
                mal_to=mal_to,
                kitsu_to=kitsu_to,
                anilist_to=anilist_to,
                episodes_to=episodes_to
            )
            instances.append(redirect)

        instance = cls(
            raw=parsed.group(0),
            mal_from=mal_from,
            kitsu_from=kitsu_from,
            anilist_from=anilist_from,
            episodes_from=episodes_from,
            mal_to=mal_to,
            kitsu_to=kitsu_to,
            anilist_to=anilist_to,
            episodes_to=episodes_to
        )

        instances.append(instance)

        return instances

    def get_episode_redirect(self, episode: int) -> Optional[int]:
        """
        Retrieves the redirect for a passed episode.

        Parameters
        ----------
        episode: int
            The episode to get the redirect
            for.

        Returns
        -------
        Optional[int]
            The redirected episode.
            None if it could not be calculated.
        """
        try:
            if episode < self.episodes_from[0] or episode > self.episodes_from[1]:
                return None
        except TypeError:
            pass

        offset = episode - self.episodes_from[0]

        try:
            if offset + self.episodes_to[0] > self.episodes_to[1]:
                return None
        except TypeError:
            pass

        return offset + self.episodes_to[0]



from setuptools import setup
import re


requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

version = ''
with open('anime_relations_py/__init__.py') as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

readme = ''
with open('README.md') as f:
    readme = f.read()

setup(
    name="anime-relations-py",
    author='Tyler',
    url="https://github.com/tylergibbs2/anime-relations-py",
    project_urls={
        "Documentation": "https://github.com/tylergibbs2/anime-relations-py/blob/master/README.md",
        "Issue tracker": "https://github.com/tylergibbs2/anime-relations-py/issues",
    },
    version=version,
    license='MIT',
    description="A parser for anime-relations. So you don't have to.",
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    python_requires='>=3.6',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ]
)
